/*==============================================================================*/
/*  Copyright (C) 2021, Heartland Data inc. All Rights Reserved.                */
/*                                                                              */
/*  Title  :   Data Writing Driver                                              */
/*  EventID:   Insert Extended Event-ID Test Point                              */
/*  FileID :   39a                                                              */
/*  Version:   1.2                                                              */
/*  Author :   HLDC                                                             */
/*==============================================================================*/

/*==============================================================================*/
/*  Desc:   Header for File IO                                                  */
/*==============================================================================*/
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/version.h>
#include <linux/proc_fs.h>
#include <asm/uaccess.h>
#include <linux/syscalls.h>
#include <linux/sched.h>
#include <linux/vmalloc.h>
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 6, 0))
#include <linux/ktime.h>
#endif

/*==============================================================================*/
/*  Macro:  DT_UINT                                                             */
/*  Desc:   Please change Test Point argument type for DT+ Project setting.     */
/*==============================================================================*/
#define	DT_UINT unsigned int

/*==============================================================================*/
/*  Desc:   Use Ring Buffer                                                     */
/*==============================================================================*/
#define RING_BUFF_USE

/*==============================================================================*/
/*  Desc:   Test Point Parametaer                                               */
/*==============================================================================*/
#define DT_VARIABLE_BIT		0x2
#define	DT_EVTTRG_BIT		0x8

/*==============================================================================*/
/*  Macro:  DT_ADD_CPU_INFO                                                     */
/*  Desc:   Please set 1 when you add CPU ID to Test Point.                     */
/*==============================================================================*/
#define DT_ADD_CPU_INFO		1

/*==============================================================================*/
/*  Desc:   Please set buffer size                                              */
/*==============================================================================*/
#define DT_MAX_TESTREPORT_NUMBER	800000	/* Customize */
#define DT_PROC_BUFF_SIZE	(DT_MAX_TESTREPORT_NUMBER * 18)

#define	PROC_NAME	"dtproc"

//static unsigned char buff[DT_PROC_BUFF_SIZE];
static unsigned char *buff = NULL;
static int write_len = 0, write_pos = 0;
static int start_pos = 0;
static int busout_mode = 0;
#ifdef RING_BUFF_USE
static int buff_state = 0; /* buff_state == 1 is buffer full state */
#endif
unsigned int cpu_id = 0;
static DEFINE_SPINLOCK(tplock);

struct DT_DATA
{
	char tp_func;
	DT_UINT addr; 
	DT_UINT dat;
	DT_UINT size;
	unsigned char value[256];
};

#define TP_BusOut 		'3'
#define TP_MemoryOutput '4'

/*==============================================================================*/
/*  Macro:  BusOutModeSet                                                       */
/*  Desc:   0 => stop logging 1 => start logging                                */
/*==============================================================================*/
void SetBusOutMode(int in_mode)
{
	busout_mode = in_mode;
}
EXPORT_SYMBOL(SetBusOutMode);

void InitBuffInfo(void)
{
	SetBusOutMode(0);
#ifdef RING_BUFF_USE
	buff_state = 0;
#endif
	write_len = 0;
	write_pos = 0;
	start_pos = 0;
	if (buff) {
		vfree(buff);
		buff = NULL;
	}
}
EXPORT_SYMBOL(InitBuffInfo);

/*==============================================================================*/
/*  Func:   _TP_BusOutDrv                                                       */
/*  Desc:   Called by Test Point                                                */
/*==============================================================================*/
static inline void _TP_BusOutDrv(DT_UINT addr, DT_UINT dat)
{
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5,6,0))
	struct timespec64 now;
#else
	struct timespec now;
#endif

	if (write_pos >= DT_PROC_BUFF_SIZE) {
#ifdef RING_BUFF_USE
		buff_state = 1;
		write_pos = 0;
#else
		return;
#endif
	}

#if DT_ADD_CPU_INFO
	cpu_id = smp_processor_id();
#endif
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 6, 0))
	ktime_get_real_ts64(&now);
#else
	getnstimeofday(&now);
#endif

	buff[write_pos++] = (unsigned char)((dat >> 0) & 0xFF);
	buff[write_pos++] = (unsigned char)((dat >> 8) & 0xFF);
	buff[write_pos++] = (unsigned char)((addr >> 0) & 0xFF);
	buff[write_pos++] = (unsigned char)((addr >> 8) & 0xFF);
	buff[write_pos++] = (unsigned char)((addr >> 16) & 0xFF);

   	/* time data */
	buff[write_pos++] = (unsigned char)((now.tv_nsec) & 0xFF);
	buff[write_pos++] = (unsigned char)((now.tv_nsec >> 8) & 0xFF);
	buff[write_pos++] = (unsigned char)((now.tv_nsec >> 16) & 0xFF);
	buff[write_pos++] = (unsigned char)((now.tv_nsec >> 24) & 0xFF);
	buff[write_pos++] = (unsigned char)((now.tv_sec) & 0xFF);
	buff[write_pos++] = (unsigned char)((now.tv_sec >> 8) & 0xFF);
	buff[write_pos++] = (unsigned char)((now.tv_sec >> 16) & 0xFF);
	buff[write_pos++] = (unsigned char)((now.tv_sec >> 24) & 0xFF);

	/* cpu id */
	buff[write_pos++] = (unsigned char)cpu_id;

	/* Reserve */
	buff[write_pos++] = 0x00;
	buff[write_pos++] = 0x00;
	buff[write_pos++] = 0x00;
	buff[write_pos++] = 0x00;

	write_len += 18;
#ifdef RING_BUFF_USE
	if (buff_state == 1) {
		start_pos = write_pos;
		write_len -= 18;
	}
#endif
}

/*==============================================================================*/
/*  Func:   _TP_MemoryOutputDrv                                                 */
/*  Desc:   Called by Variable Test Point                                       */
/*==============================================================================*/
static inline void _TP_MemoryOutputDrv(unsigned char *p, DT_UINT size)
{
	int	i, packetNum, count = 0;

	if( size >= 256 ) size = 256;
	packetNum = (size + 7) / 8;

	for (i = 0; i < packetNum; i++) {
		if (write_pos >= DT_PROC_BUFF_SIZE) {
#ifdef RING_BUFF_USE
			buff_state = 1;
			write_pos = 0;
#else
			return;
#endif
		}

		buff[write_pos++] = size;
		buff[write_pos++] = packetNum;
		buff[write_pos++] = 0x03;
		buff[write_pos++] = i;
		buff[write_pos++] = (count < size)?*p:0x00;
		count++; p++;
		buff[write_pos++] = (count < size)?*p:0x00;
		count++; p++;
		buff[write_pos++] = (count < size)?*p:0x00;
		count++; p++;
		buff[write_pos++] = (count < size)?*p:0x00;
		count++; p++;
		buff[write_pos++] = (count < size)?*p:0x00;
		count++; p++;
		buff[write_pos++] = (count < size)?*p:0x00;
		count++; p++;
		buff[write_pos++] = (count < size)?*p:0x00;
		count++; p++;
		buff[write_pos++] = (count < size)?*p:0x00;
		count++; p++;
		buff[write_pos++] = 0x00;
		buff[write_pos++] = 0x00;
		buff[write_pos++] = 0x00;
		buff[write_pos++] = 0x00;
		buff[write_pos++] = 0x00;
		buff[write_pos++] = 0x00;
		write_len += 18;
#ifdef RING_BUFF_USE
		if (buff_state == 1) {
			start_pos = write_pos;
			write_len -= 18;
		}
#endif
	}
}

/*==============================================================================*/
/*  Func:   _TP_BusOut                                                          */
/*  Desc:   Called by Test Point                                                */
/*==============================================================================*/
void _TP_BusOut(DT_UINT addr, DT_UINT dat)
{
	unsigned long flags;

	if (busout_mode != 1) {
		return;
	}

	spin_lock_irqsave(&tplock, flags);
	_TP_BusOutDrv(addr, dat);
   	spin_unlock_irqrestore(&tplock, flags);
}
EXPORT_SYMBOL(_TP_BusOut);

/*==============================================================================*/
/*  Func:   _TP_MemoryOutput                                                    */
/*  Desc:   Called by Variable TestPoint                                        */
/*==============================================================================*/
void _TP_MemoryOutput(DT_UINT addr, DT_UINT dat, void *value, DT_UINT size)
{
	unsigned long flags;

	if (busout_mode != 1) {
		return;
	}

	spin_lock_irqsave(&tplock, flags);
	_TP_BusOutDrv( addr|DT_VARIABLE_BIT, dat );		//Memory Dump Start
	_TP_MemoryOutputDrv((unsigned char*)value, size);
	spin_unlock_irqrestore(&tplock, flags);
}
EXPORT_SYMBOL(_TP_MemoryOutput);

/*==============================================================================*/
/*  Func:   _TP_EventTrigger32                                                  */
/*  Desc:   Called by Event Trigger 32bit Test Point                            */
/*==============================================================================*/
void _TP_EventTrigger32( DT_UINT addr, DT_UINT dat, DT_UINT event_id )
{
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5,6,0))
	struct timespec64 now;
#else
	struct timespec now;
#endif
	unsigned long flags;

	if (busout_mode != 1) {
		return;
	}

	if (write_pos >= DT_PROC_BUFF_SIZE) {
#ifdef RING_BUFF_USE
		buff_state = 1;
		write_pos = 0;
#else
		return;
#endif
	}

	spin_lock_irqsave(&tplock, flags);
#if DT_ADD_CPU_INFO
	cpu_id = smp_processor_id();
#endif
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 6, 0))
	ktime_get_real_ts64(&now);
#else
	getnstimeofday(&now);
#endif

	addr |= DT_EVTTRG_BIT;
	buff[write_pos++] = ((dat >> 0) & 0xFF);
	buff[write_pos++] = ((dat >> 8) & 0xFF);
	buff[write_pos++] = ((addr >> 0) & 0xFF);
	buff[write_pos++] = ((addr >> 8) & 0xFF);
	buff[write_pos++] = ((addr >> 16) & 0xFF);

   	/* time data */
	buff[write_pos++] = (unsigned char)((now.tv_nsec) & 0xFF);
	buff[write_pos++] = (unsigned char)((now.tv_nsec >> 8) & 0xFF);
	buff[write_pos++] = (unsigned char)((now.tv_nsec >> 16) & 0xFF);
	buff[write_pos++] = (unsigned char)((now.tv_nsec >> 24) & 0xFF);
	buff[write_pos++] = (unsigned char)((now.tv_sec) & 0xFF);
	buff[write_pos++] = (unsigned char)((now.tv_sec >> 8) & 0xFF);
	buff[write_pos++] = (unsigned char)((now.tv_sec >> 16) & 0xFF);
	buff[write_pos++] = (unsigned char)((now.tv_sec >> 24) & 0xFF);

	/* cpu id */
	buff[write_pos++] = (unsigned char)cpu_id;

	/* Event ID */
	buff[write_pos++] = (unsigned char)((event_id) & 0xFF);
	buff[write_pos++] = (unsigned char)((event_id >> 8) & 0xFF);
	buff[write_pos++] = (unsigned char)((event_id >> 16) & 0xFF);
	buff[write_pos++] = (unsigned char)((event_id >> 24) & 0xFF);
	write_len += 18;
#ifdef RING_BUFF_USE
	if (buff_state == 1) {
		start_pos = write_pos;
		write_len -= 18;
	}
#endif
   	spin_unlock_irqrestore(&tplock, flags);
}
EXPORT_SYMBOL(_TP_EventTrigger32);

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(3, 4, 0))
static ssize_t dt_proc_write(struct file *filp, const char __user *buf, size_t len, loff_t *data)
#else
static int dt_proc_write(struct file *filp, const char *buf, unsigned long len, void *data)
#endif
{
	struct DT_DATA dt_data;
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 4, 0))
	char tmp;

	if (copy_from_user(&tmp, buf, 1)) {
		return len;
	}
	switch (tmp) {
#else
	switch (buf[0]) {
#endif
	case TP_BusOut :
		if(copy_from_user(&dt_data, buf, len))
			return -EFAULT;
		_TP_BusOut(dt_data.addr, dt_data.dat);
		break;
	case TP_MemoryOutput :
		if(copy_from_user(&dt_data, buf, len))
			return -EFAULT;
		_TP_MemoryOutput(dt_data.addr, dt_data.dat, dt_data.value, dt_data.size);
		break;
	case '0' :
		SetBusOutMode(0);
		printk( KERN_INFO "[DT] Trace Stop.\n" );
		break;
	case '1' :
		buff = vmalloc(DT_PROC_BUFF_SIZE);
		if( buff == NULL ){
			printk(KERN_INFO "DT Cannot alloc vmalloc\n");
			return len;
		}
		SetBusOutMode(1);
		printk( KERN_INFO "[DT] Trace Start.\n" );
		break;
	default:
		InitBuffInfo();
		printk( KERN_INFO "[DT] Buffer Clear.\n");
		break;
	};

	return len;
}

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(3, 4, 0))
static ssize_t dt_proc_read(struct file *filp, char __user *buf, size_t count, loff_t *data)
#else
static int dt_proc_read(char *buf, char **start, off_t off, int count, int *eof, void *data)
#endif
{
	int read_size = 0;
	int i;

	SetBusOutMode(0);
	if (write_len <= 0) {
#if !(LINUX_VERSION_CODE >= KERNEL_VERSION(3, 4, 0))
		*eof = 1;
#endif
		InitBuffInfo();
		return 0;
	}

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 4, 0))
#ifdef RING_BUFF_USE
	if (count < write_len) {
		if ((start_pos + count) > DT_PROC_BUFF_SIZE) {
			read_size = DT_PROC_BUFF_SIZE - start_pos;
			if (copy_to_user(buf, (buff + start_pos), read_size)) {
				return 0;
			}
			start_pos = 0;
		} else {
			read_size = count;
			if (copy_to_user(buf, (buff + start_pos), read_size)) {
				return count;
			}
			start_pos += read_size;
		}
		write_len -= read_size;
	} else {
		if ((start_pos + write_len) > DT_PROC_BUFF_SIZE) {
			read_size = DT_PROC_BUFF_SIZE - start_pos;
			if (copy_to_user(buf, (buff + start_pos), read_size)) {
				return 0;
			}
			start_pos = 0;
		} else {
			read_size = write_len;
			if (copy_to_user(buf, (buff + start_pos), read_size)) {
				return 0;
			}
			start_pos += read_size;
		}
		write_len -= read_size;
	}
#else
	if (count < write_len) {
		read_size = count;
	} else {
		read_size = write_len;
	}
	if (copy_to_user(buf, (buff + start_pos), read_size)) {
		return 0;
	}
	start_pos += read_size;
	write_len -= read_size;
#endif
#else
	for (i = 0; ((i < count) && (write_len > 0)); i++) {
		buf[i] = buff[start_pos++];
#ifdef RING_BUFF_USE
		if (start_pos == DT_PROC_BUFF_SIZE) {
			start_pos = 0;
		}
#endif
		write_len--;
	}

#if !(LINUX_VERSION_CODE >= KERNEL_VERSION(3, 4, 0))
	*start = buf;
#endif
	read_size = i;
#endif

	return read_size;
}

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 6, 0))
static const struct proc_ops dt_proc_fops = {
	.proc_write = dt_proc_write,
	.proc_read = dt_proc_read,
};
#elif (LINUX_VERSION_CODE >= KERNEL_VERSION(3, 4, 0))
static const struct file_operations dt_proc_fops = {
	.write = dt_proc_write,
	.read = dt_proc_read,
};
#endif

static int __init dt_proc_init(void)
{
	struct proc_dir_entry* entry;

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(3, 4, 0))
	entry = proc_create(PROC_NAME, S_IRUGO | S_IWUGO, NULL, &dt_proc_fops);
#else
	entry = create_proc_entry(PROC_NAME, S_IRUGO | S_IWUGO, NULL);
#endif
	if ( !entry ) {
		printk( KERN_ERR "proc entry fail\n" );
		return -EBUSY;
	}
#if !(LINUX_VERSION_CODE >= KERNEL_VERSION(3, 4, 0))
	entry->read_proc = dt_proc_read;
	entry->write_proc = dt_proc_write;
#endif
	printk(KERN_INFO "dtproc load OK\n");

	return 0;
}

static void __exit dt_proc_exit(void)
{
	remove_proc_entry(PROC_NAME, NULL);
	printk(KERN_INFO "dtproc unload\n");
}

module_init(dt_proc_init);
module_exit(dt_proc_exit);

MODULE_DESCRIPTION("dtproc");
MODULE_LICENSE("GPL2");
